--[[

OverView
	This is simply an interface built ontop of oUF.
	Instead of creating loads of spawn functions,
	editing, and checks, this file takes a table as
	input, and spawns the apropriate unitframes.
	
	For API, use oUF_HolySmurf as an example.

PlayerTargets
	Besides that, it handles Raid-PlayerTargets
	/pt <xxx>	Sets a playertarget
	/clearpt	Clears the PT list
	
Extras
	Another extra feature is barEnhancement,
	by adding a gradient texture to the end of
	statusbars, increasing the contrast with the
	background.
	For more info, look at
	    element.health.enhanceXXX 
	    in:   oUF_HolySmurf/templates.lua

]]--
if not oUF then return end

oUF.Smurf = { }
local Smurf = oUF.Smurf
local config = {}
local print = function(msg) DEFAULT_CHAT_FRAME:AddMessage(msg) end
local party, playertarget -- Pointers for hiding and adding units

--------------------------------------
-- 	Helper functions
--------------------------------------
function Smurf.copyTable(from)
	local to = {}
	for k,v in pairs(from) do
		if type(v) == "table" then
			to[k] = {}
			to[k] = Smurf.copyTable(from[k])
		else
			to[k] = v
		end
	end
	return to
end

local cfg = function(self, ...)
	if type(select(1, ...)) == "function" then
		local func = select(1, ...)
		return func(self)
	else
		return ...
	end
end

local getValues = function(self, db, ... )
	local t = { }
	for i=1, select('#', ...) do
		table.insert(t, cfg(self, db[select(i, ...)]))
	end
	return unpack(t)
end

local updateSparkPosHorizontal = function(self, event, unit, bar, min, max)
	local x = bar:GetWidth()*(min/max)-bar.stripe:GetWidth()
	if min == max or x < 0 then
		bar.stripe:Hide()
	else
		bar.stripe:SetPoint("LEFT", bar, "LEFT", x, 0)
		if not bar.stripe:IsShown() then
			 bar.stripe:Show()
		end
	end
end

local updateSparkPosVertical = function(self, event, unit, bar, min, max)
	local y = bar:GetHeight()*(min/max)-bar.stripe:GetHeight()
	if min == max or y < 0 then
		bar.stripe:Hide()
	else
		bar.stripe:SetPoint("BOTTOM", bar, "BOTTOM", 0, y)
		if not bar.stripe:IsShown() then
			 bar.stripe:Show()
		end
	end
end

local stdBar = function(self, element, unit)
	local f = CreateFrame("StatusBar", nil, self)
	f:SetStatusBarTexture( cfg(self, element.texture) )
	local orientation = cfg(self, element.orientation)
	if element.enhanceEnable then
		f.stripe = self.high:CreateTexture(nil, "OVERLAY")
		f.stripe:SetTexture( cfg(self, element.enhanceTexture) )
		f.stripe:SetBlendMode("ADD")
		f.stripe:SetAlpha( cfg(self, element.enhanceAlpha) )
		if orientation == "HORIZONTAL" then
			f.stripe:SetPoint("TOP", f, "TOP")
			f.stripe:SetPoint("BOTTOM", f, "BOTTOM")
			f.stripe.update = updateSparkPosHorizontal
			f.stripe:SetWidth( cfg(self, element.enhanceSize) )
		else
			f.stripe:SetPoint("LEFT", f, "LEFT")
			f.stripe:SetPoint("RIGHT", f, "RIGHT")
			f.stripe.update = updateSparkPosVertical
			f.stripe:SetTexCoord( 1, 1,
								  0, 1,
								  1, 0,
								  0, 0 )
			f.stripe:SetHeight( cfg(self, element.enhanceSize) )
		end
	end
	f:SetOrientation( orientation )
	-- oUF coloring
	f.colorClass = cfg(self, element.colorClass)
	f.colorTapping = cfg(self, element.colorTapping)
	f.colorReaction = cfg(self, element.colorReaction)
	f.colorClassNPC = cfg(self, element.colorClassNPC)
	f.colorHappiness = cfg(self, element.colorHappiness)
	f.colorDisconnected = cfg(self, element.colorDisconnected)
	return f
end

local stdFont = function(self, element)
	local fs = self.high:CreateFontString(nil, "OVERLAY")
	fs:SetFont(getValues(self, element, "font", "fontSize", "fontOutline"))
	if element.fontColor then fs:SetTextColor( cfg(self, element.fontColor ) ) else fs:SetTextColor(1,1,1) end
	if element.fontJustifyH then fs:SetJustifyH( cfg(self, element.fontJustifyH ) ) end
	if element.fontJustifyV then fs:SetJustifyV( cfg(self, element.fontJustifyV ) ) end
	if element.fontShadowColor then fs:SetShadowColor( cfg(self, element.fontShadowColor ) ) else fs:SetShadowColor(0, 0, 0) end
	if element.fontShadowOffset then fs:SetShadowOffset( cfg(self, element.fontShadowOffset ) ) else fs:SetShadowOffset(0.8, -0.8) end
	return fs
end

--------------------------------------
-- 	Element functions
--------------------------------------

Smurf.health = function(self, element, unit)
	self.Health = stdBar(self, element, unit)
	if element.enhanceEnable then
		local old = self.PostUpdateHealth
		self.PostUpdateHealth = function(...)
			self.Health.stripe.update(...)
			if old then old(...) end
		end
	end	
	return self.Health
end

Smurf.healthtext = function(self, element, unit)
	self.HealthText = stdFont(self, element)
	local old = self.PostUpdateHealth
	self.PostUpdateHealth = function(...)
		element.onRefresh(...)
		if old then old(...) end
	end
	return self.HealthText
end

Smurf.power = function(self, element, unit)
	self.Power = stdBar(self, element, unit)
	if element.enhanceEnable then
		local old = self.PostUpdateHealth
		self.PostUpdatePower = function(...)
			self.Power.stripe.update(...)
			if old then old(...) end
		end
	end	
	self.Power.colorPower = cfg(self, element.colorPower)
	return self.Power
end

Smurf.powertext = function(self, element, unit)
	self.PowerText = stdFont(self, element)
	local old = self.PostUpdatePower
	self.PostUpdatePower = function(...)
		element.onRefresh(...)
		if old then old(...) end
	end
	return self.PowerText
end

Smurf.druidmana = function(self, element, unit)
	local _, class = UnitClass("player")
	if(class == "DRUID") then		
		self.DruidMana = stdBar(self, element, unit)
	end
	return self.DruidMana
end

Smurf.druidmanatext = function(self, element, unit)
	local _, class = UnitClass("player")
	if(class == "DRUID") then		
		self.DruidMana.text = stdFont(self, element)
	end
	return self.DruidMana.text
end

Smurf.name = function(self, element, unit)
	local name = stdFont(self, element)
	self.Name = name
	if type(element.onRefresh) == "function" then
		self.UNIT_NAME_UPDATE = element.onRefresh
	end
	return self.Name
end

Smurf.highlight = function(self, element, unit)
	self.Highlight = self.high:CreateTexture(nil, "OVERLAY")
	self.Highlight:SetTexture( cfg(self, element.texture) )
	self.Highlight:SetBlendMode("ADD")
	self.Highlight:SetAllPoints(self)
	self.Highlight:SetAlpha( cfg(self, element.alpha) )
end

Smurf.background = function(self, element, unit)
	local padding = element.padding or UIParent:GetEffectiveScale() -- 1 px
	self.Background = self:CreateTexture(nil, "BACKGROUND")
	self.Background:SetPoint("LEFT", self, "LEFT", -padding, 0)
	self.Background:SetPoint("RIGHT", self, "RIGHT", padding, 0)
	self.Background:SetPoint("TOP", self, "TOP", 0, padding)
	self.Background:SetPoint("BOTTOM", self, "BOTTOM", 0, -padding)
	self.Background:SetTexture( cfg(self, element.texture) )
	self.Background:SetVertexColor( cfg(self, element.color) )
	self.Background:SetAlpha( cfg(self, element.alpha) )
end

Smurf.raidicon = function(self, element, unit)
	self.RaidIcon = self.high:CreateTexture(nil, "OVERLAY")
	self.RaidIcon:SetTexture("Interface\\TargetingFrame\\UI-RaidTargetingIcons")
	return self.RaidIcon
end

Smurf.leader = function(self, element, unit)
	self.Leader = self.high:CreateTexture(nil, "OVERLAY")
	self.Leader:SetTexture('Interface\\GroupFrame\\UI-Group-LeaderIcon')
	return self.Leader
end

Smurf.rangecheck = function(self, element, unit)
	self.Range = true
	self.inRangeAlpha = cfg(self, element.inRangeAlpha) or 1
	self.outsideRangeAlpha = cfg(self, element.outOfRangeAlpha) or .3
end

Smurf.combat = function(self, element, unit)
	self.Combat = self.high:CreateTexture(nil, "OVERLAY")
	self.Combat:SetTexture("Interface\\CharacterFrame\\UI-StateIcon")
	self.Combat:SetTexCoord(0.58, 0.90, 0.08, 0.41)
end

--------------------------------------
-- 	Rightclick menu
--------------------------------------
local showMenu = function(frame)
	HideDropDownMenu(1)
	ToggleDropDownMenu(1, nil, frame, "cursor")
end

local partyMenu = function(num)
	showMenu( _G["PartyMemberFrame"..num.."DropDown"] )
end

local raidMenu = function(num)
	local frame = FriendsDropDown
	frame.displayMode = "MENU"
	frame.initialize = RaidFrameDropDown_Initialize
	showMenu( frame )
end

local menu = function(self)	
	local unit = SecureButton_GetUnit(self)
	
	if UnitIsUnit(unit, "player") then
		showMenu( PlayerFrameDropDown )
		return
	elseif UnitIsUnit(unit, "pet") then
		showMenu( PetFrameDropDown )
		return
	elseif unit == "target" then
		showMenu( TargetFrameDropDown )
		return
	else
		local num = unit:match"^party(%d)$"
		if num then
			partyMenu( num )
			return
		else
			num = unit:match"^raid(%d%d?)$"
			if num then
				raidMenu( num )
				return
			end
		end
	end

	-- One desperate last look for a match..
	for i=1, 4 do
		if UnitIsUnit(unit, "party"..i) then
			partyMenu( i )
			return
		end
	end
	for i=1, 40 do
		if UnitIsUnit(unit, "raid"..i) then
			raidMenu( i )
			return
		end
	end

end

--------------------------------------
-- 	Generic style function
--------------------------------------
local smurfStyle = function(self, unit)
	if not unit then
		_, unit = strsplit("_", self:GetParent():GetName())
		self.groupUnit = true
	end
	self.unit = unit
	--print("Spawning: "..unit)
	
	-- Standard stuff
	self.menu = menu												-- enable menus
	self:SetScript("OnEnter", UnitFrame_OnEnter)
    self:SetScript("OnLeave", UnitFrame_OnLeave)
	self:RegisterForClicks("anyup") 								-- enable mouse click intercept
	self:SetAttribute("*type2", "menu")
	
	self.high = CreateFrame("Frame", nil, self)
	self.high:SetFrameLevel(self:GetFrameLevel() + 1)
	
	-- Spawn elements
	self.db = config.units[unit]
	for elementName,element in pairs(self.db.element) do
		if type(element) == "function" then
			element(self, unit)
		else
			-- Specialized stuff
			if Smurf[elementName] then
				local e = Smurf[elementName](self, element, unit)
				if e then -- General stuff
					if cfg(self, element.anchorTo) then
						e:SetPoint( getValues(self, element, "anchorPoint", "anchorTo", "anchorRelative", "anchorOffsetX", "anchorOffsetY" ) )
					end
					local w, h = cfg(self, element.width), cfg(self, element.height)
					if w then e:SetWidth(w) end
					if h then e:SetHeight(h) end
				end
			end
			if element.postFunction then
				element.postFunction(self)
			end
		end
	end
	
	self:SetAttribute('initial-width', cfg(self, self.db.width))
	self:SetAttribute('initial-height', cfg(self, self.db.height))
	
	self:Show()
	
	return self
end

--------------------------------------
-- 	Spawn pet/target frames after
--	combat ends.
--------------------------------------
local tasks = {}
local spawnHandler = CreateFrame("Frame")
spawnHandler:RegisterEvent("PLAYER_REGEN_ENABLED")
spawnHandler:SetScript("OnEvent", function()
	for key,func in pairs(tasks) do
		tasks[key]()
		tasks[key] = nil
	end
	spawnHandler:Hide()
end)
spawnHandler:Hide()

--------------------------------------
-- 	Spawn functions
--------------------------------------
local SpawnRelative
SpawnRelative = function(parent, unit, child)
	if not config.units[unit..child].spawnFromParent then return end
	if InCombatLockdown() then
		tinsert(tasks, function()
				SpawnRelative(parent, unit, child)
			end)
		return
	end
	local object = oUF:Spawn(unit..child)
	UnregisterUnitWatch(object)
	object:SetParent(parent)
	object:SetPoint( getValues(object, config.units[unit..child], "anchorPoint", "anchorTo", "anchorRelative", "anchorOffsetX", "anchorOffsetY" ) )
	object:SetAttribute("unit", nil)
	object:SetAttribute("useparent-unit", true)
	object:SetAttribute("unitsuffix", child)
	object:SetAttribute("*type1","target")
	RegisterUnitWatch(object)
	
	object:SetScript("OnShow", function() -- Tiny hack to update names etc.
		object.unit = SecureButton_GetUnit(object)
		object:PLAYER_ENTERING_WORLD()
	end )
end

local InitCallback = function(self, unit)
	if not unit then
		_, unit = strsplit("_", self:GetParent():GetName())
	end
	if not unit then return end
	
	if config.units[unit.."target"] then
		SpawnRelative(self, unit, "target")
	end
	
	if config.units[unit.."pet"] then
		SpawnRelative(self, unit, "pet")
	end
end

oUF:RegisterInitCallback(InitCallback)

oUF:RegisterSubTypeMapping('UNIT_LEVEL')
oUF:RegisterStyle('Smurf', smurfStyle)
oUF:SetActiveStyle('Smurf')

-- Spawn all single units
Smurf.Spawn = function(newConfig)
	config = newConfig
	for unit,db in pairs(config.units) do
	 	if not db.spawnFromParent and not(unit:match"raid" or unit:match"main" or unit:match"party") then
			local object = oUF:Spawn(unit)
			object:SetPoint( getValues(object, db, "anchorPoint", "anchorTo", "anchorRelative", "anchorOffsetX", "anchorOffsetY" ) )
		end
	end

	for _,db in ipairs(config.groups) do
		local group = db.name
		local object = oUF:Spawn("header", "oUFSmurf"..group.."_"..db.children)
		object:SetManyAttributes( cfg(self, db.manyAttributes) )
		object:SetPoint( getValues(object, db, "anchorPoint", "anchorTo", "anchorRelative", "anchorOffsetX", "anchorOffsetY" ) )
		object:Show()
		if group == "party" then party = object end
		if group == "raidplayertarget" then playertarget = object end
	end
end

----------------------------------
--	Partyhandling	--
----------------------------------
if not config.showPartyInRaid then
	local e = CreateFrame("Frame")
	e:RegisterEvent("PARTY_MEMBERS_CHANGED")
	e:RegisterEvent("PARTY_LEADER_CHANGED")
	e:RegisterEvent("RAID_ROSTER_UPDATE")
	e:RegisterEvent("PLAYER_LOGIN")
	e:SetScript("OnEvent", function(self)
		if not party then return end
		if(InCombatLockdown()) then
			self:RegisterEvent("PLAYER_REGEN_ENABLED")
		else
			self:UnregisterEvent("PLAYER_REGEN_ENABLED")
			if(GetNumRaidMembers() > 0) then
				party:SetAttribute("showRaid", true)
				party:SetAttribute("showParty", false)
				party:Hide()
			elseif(GetNumPartyMembers() > 0) then
				party:SetAttribute("showRaid", false)
				party:SetAttribute("showParty", true)
				party:Show()
			end
		end
	end)
end

----------------------------------
--	PlayerTargets	--
----------------------------------
local PlayerTargets = {}
local SetPlayerTarget = function(unit)
	if not playertarget then return end
	local unitName, realm = (unit and UnitName(unit)) or UnitName("target")
	if realm then unitName = unitName.."-"..realm end
	if unitName then
		for _,pt in pairs(PlayerTargets) do
			if pt == unitName then
				print("|cFFFFFF00"..pt.." is already in the playertarget list.")
				return
			end
		end
		tinsert(PlayerTargets, unitName)
		playertarget:SetAttribute("nameList", strjoin(",", unpack(PlayerTargets)))
	else
		print("|cFFFFFF00Please select a target or suply a valid raid/party member.")
	end
end

local ClearPlayerTarget = function()
	if not playertarget then return end
	PlayerTargets = {}
	playertarget:SetAttribute("nameList", "")
end

SLASH_SETPLAYERTARGET1 = "/playertarget"
SLASH_SETPLAYERTARGET2 = "/pt"
SlashCmdList["SETPLAYERTARGET"] = SetPlayerTarget;

SLASH_CLEARPLAYERTARGET1 = "/clearplayertarget"
SLASH_CLEARPLAYERTARGET2 = "/clearpt"
SlashCmdList["CLEARPLAYERTARGET"] = ClearPlayerTarget;

----------------------------------------------------------------------------
-- 		Hook into OmniCC if exists
----------------------------------------------------------------------------
if OmniCC then
	local oldCreateTimer = OmniCC.CreateTimer
	OmniCC.CreateTimer = function(self, cooldown)
		local timer = oldCreateTimer(self, cooldown)
		local PostCreate = cooldown.PostCreateCooldown
		if type(PostCreate) == "function" then
			PostCreate(timer)
		end
		return timer
	end
end
